"""
$RCSfile: FourSuite12Processor.py,v $

This class encapsulates an XSLT Processor for use by XMLTransform.
This is the 4Suite 0.12 alpha 2 version, including support for XSLT
parameters and URL/URN resolution.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/11/01 02:12:24 $'[6:-2]
__version__ = '$Revision: 1.18 $'[10:-2]

# 4Suite
from Ft.Xml import XPath
from Ft.Xml.InputSource import InputSource
from Ft.Xml.Xslt import XsltException
from Ft.Xml.Xslt.Processor import Processor

# Zope builtins
from Acquisition import aq_get

# python
from cStringIO import StringIO
import os.path, sys, types, traceback

# local peer classes
from IXSLTProcessor import IXSLTProcessor

################################################################
# Defaults
################################################################

namespacesPropertyName        = 'URNnamespaces'
parametersPropertyName        = 'XSLparameters'

################################################################
# FourSuite12Processor class
################################################################

class FourSuite12Processor:
    """
    
    This class encapsulates an XSLT Processor for use by XMLTransform.
    This is the 4Suite 0.12 version, including support for XSLT
    parameters and URL/URN resolution.

    """

    __implements__ = IXSLTProcessor
    name           = '4Suite-0.12.0a3'
        
    def __init__(self):
        "Initialize a new instance of FourSuite12Processor"
        self.debugLevel = 0

    ################################################################
    # Methods implementing the IXSLTProcessor interface
    ################################################################

    def setDebugLevel(self, level):
        """

        Set debug level from 0 to 3.
        0 = silent
        3 = extra verbose
        Debug messages go to Zope server log.

        """
        self.debugLevel   = level

    def transform(self, xmlContents, xmlURL, xsltContents, xsltURL,
                  transformObject = None, REQUEST = None):
        """

        Transforms the passed in XML into the required output (usually
        HTML) using the passed in XSLT.  Both the XML and XSLT strings
        should be well-formed.  Returns the output as a string.
        transformObject and REQUEST params may be used to acquire Zope
        content such as XSLT parameters and URN namespaces, if
        required.  Catches any exceptions thrown by transformGuts and
        sends the error output to stderr, returns empty string to the
        caller.  The idea is that web site users will at worst see an
        empty page.

        """
        topLevelParams = None        
        if transformObject is not None:
            topLevelParams = self.getXSLParameters(transformObject)            
        if self.debugLevel > 1:
            print "params:", topLevelParams

        if self.debugLevel > 1:
            print "xsltContents:"
            print xsltContents
            print "xmlContents:"
            print xmlContents

        try:
            result = self.transformGuts(xmlContents, xmlURL, xsltContents, xsltURL,
                                        transformObject, topLevelParams, REQUEST)

        except Exception, e:
            sys.stderr.write(str(e) + '\n')
            return ""
            
        return result

    def addParam(self, paramMap, name, value):
        """

        This is a convenience function for adding parameters in the
        correct format to the parameter map to be used for the
        'params' parameter in transformGuts.
        
        """
        paramMap[ (None, name) ] = value
        return paramMap
    
    def transformGuts(self, xmlContents, xmlURL, xsltContents, xsltURL,
                      transformObject = None, params = None,
                      REQUEST = None):
        """

        Actually performs the transformation.  Throws an Exception if
        there are any errors.
        
        """
        namespaceMap   = {}
        if transformObject is not None:
            namespaceMap   = self.retrieveNamespaces(transformObject)
        if self.debugLevel > 1:
            print "namespaces:", namespaceMap

        if self.debugLevel > 1:
            from Ft.Xml.Xslt import ExtendedProcessingElements
            processor = ExtendedProcessingElements.ExtendedProcessor()
            if self.debugLevel > 2:
                processor._4xslt_debug = 1
                processor._4xslt_trace = 1
        else:
            processor = Processor()

        try:
            docSrc   = ResolvingInputSource(namespaceMap, REQUEST, 
                                            StringIO(xmlContents),
                                            xmlURL)
            styleSrc = ResolvingInputSource(namespaceMap, REQUEST,
                                            StringIO(xsltContents),
                                            xsltURL)
            processor.appendStylesheet(styleSrc)
            result = processor.run(docSrc, topLevelParams = params)

            if self.debugLevel > 1:
                print "===Result==="
                print result
                print "============"

        except XsltException, e:
            #(ty, val, tb) = sys.exc_info()
            #traceback.print_tb(tb)
            raise Exception(str(e))
            
        except (XPath.RuntimeException, XPath.CompiletimeException), e:
            if hasattr(e, 'stylesheetUri'):
                message = "While processing %s\n%s" % e.stylesheetUri, str(e)
                raise Exception(message)
            else:
                raise Exception(str(e))

        return result

    ################################################################
    # Internal methods
    ################################################################

    def retrieveNamespaces(self, transformObject):
        """

        retrieves Namespaces defined for URI Resolution

        """
        NIDs   = aq_get(transformObject,namespacesPropertyName,None)
        result = {}
        if NIDs is not None:
            for n in NIDs:
                value = aq_get(transformObject,n,None)
                # I use callable() to determine if it is not a scalar.
                # If not, it must be a Zope object (I think) - WGM
                if callable(value):
                    result[n] = value
                else:
                    result[n] = str(value)
        return result

    def getXSLParameters(self, transformObject):
        """

        Return XSL Transformer parameters as a dictionary of strings
        in the form 'name:value' as would be passed to an XSLT engine
        like Saxon, 4suite, etc. The values are obtained by looking
        for a property in the current context called 'XSLparameters',
        which should be a list of strings. Each name on the list is
        looked up in the current context. If its value is a scalar,
        then the pair 'name:value' is returned. If the value is an
        object, then the pair 'name:url' is returned where url is the
        absolute URL of the object.  The key (name) is actually a
        tuple of two strings, the first of which is an optional
        namespace (we don't use this today).

        """
        parms  = aq_get(transformObject,parametersPropertyName,None)
        result = {}
        if parms is not None:
            for p in parms:
                value = aq_get(transformObject,p,None)
                # I use callable() to determine if it is not a scalar.
                # If not, it must be a Zope object (I think) - WGM
                if callable(value):
                    self.addParam(result, p, value.absolute_url())
                else:
                    self.addParam(result, p, str(value))
        return result

################################################################
# ResolvingInputSource class
################################################################
    
class ResolvingInputSource(InputSource):
    "An input source that resolves URNs to Zope objects"

    def __init__(self, namespaceMap, REQUEST, *args, **kwargs):
        """

        Remember the URN namespaces corresponding to Zope folders and
        the REQUEST context with which we want to load the resources

        """
        apply(InputSource.__init__, (self,) + args, kwargs)
        self.namespaceMap = namespaceMap
        self.req          = REQUEST
      
    def acquireObjectContents(self, base, contextURL, REQUEST):
        """

        Obtain the contents of the Zope object indicated by the passed
        in context, starting from the passed in base object.

        """
        #print "acquire contents for:",contextURL
        zObject     = base
        #print "base", zObject.getId()
        #
        # why doesn't the below work?  Is this a bug?
        # (see com/arielpartners/website/scripts/resolver.dtml)
        # zObject = base.restrictedTraverse(contextURL)
        
        # sigh.   Do it the hard way.
        contextList = contextURL.split('/')
        for context in contextList:
            zObject = aq_get(zObject,context,None)
            if zObject is None:
                return None, None
        contents = zObject(zObject, REQUEST)
        return (contents, zObject.absolute_url())

    def resolve(self, sysid, pubid=None, hint=None):
        "Resolve the URN to a Zope object, or pass on the request"
        uriParts = sysid.split(':')
        if uriParts[0] == 'urn' and len(uriParts) == 3:
            base = self.namespaceMap.get(uriParts[1], None)
            if base is None:
                # revert to normal behavior
                return InputSource.resolve(self, sysid, pubid, hint)
            elif type(base) == types.StringType:
                # We are mapping one URL to another a la XMLCatalog RewriteURI
                #
                # could use urllib join, but it replaces the last component if no trailing slash. e.g.
                #
                # urllib.join  ("http://www.foo.com/bar", "mumble.xml") ==> http://www.foo.com/mumble.xml
                # os.path.join ("http://www.foo.com/bar", "mumble.xml") ==> http://www.foo.com/bar/mumble.xml
                resolvedURL = os.path.join(base,uriParts[2])
                return InputSource.resolve(self, resolvedURL, pubid, hint)
            else: # its a Zope object, we must retrieve its contents
                st, resolvedURL = self.acquireObjectContents(base, uriParts[2], self.req)
                if st is None:
                    # failure, cannot grab object, revert to normal behavior
                    return InputSource.resolve(self, sysid, pubid, hint)
                else:
                    # load the resource from the Zope object
                    stream = StringIO(st)
                    return self._clone(stream, resolvedURL)                    
        else:
            # revert to normal behavior
            return InputSource.resolve(self, sysid, pubid, hint)

    def _clone(self, stream, uri):
        "Override the cloning of this input source to account for the extra params."
        return self.__class__(self.namespaceMap, self.req,
                              stream, uri,
                              processIncludes=self.processIncludes,
                              stripElements=self.stripElements,
                              factory=self.factory)
